import java.math.*;
import java.util.*;

public class ModSqrt
{
  private static BigInteger ONE = BigInteger.ONE;
  private static BigInteger ZERO = BigInteger.ZERO;
  private static BigInteger TWO = new BigInteger("2");
  private static BigInteger FOUR = new BigInteger("4");
  private static Random rand = new Random();

  public static BigInteger modSqrt(BigInteger n, BigInteger p)
  {
    if (p.equals(TWO)) {
      if (n.mod(p).equals(ZERO)) {
        return ZERO;
      }
      return ONE;
    }
    if (n.mod(p).equals(ZERO)) return ZERO;
    if (!isQR(n, p)) return null;
    BigInteger pSubOne = p.subtract(ONE);
    int s = 0;
    while (!pSubOne.testBit(0)) {
      pSubOne = pSubOne.divide(TWO);
      s++;
    }
    if (s == 1) return modPow(n, (p.add(ONE)).divide(FOUR), p);
    BigInteger q = pSubOne;
    BigInteger z = selectZ(p);
    //System.out.println("Mod Square Root Z: " + z);
    BigInteger c = modPow(z, q, p);
    BigInteger r = modPow(n, (q.add(ONE)).divide(TWO), p);
    BigInteger t = modPow(n, q, p);
    int m = s;
    BigInteger b, tmp;
    int i, j;
    while(!t.equals(ONE)) {
      //System.out.println("Mod Square Root t:" + t);
      i = 0;
      tmp = t.mod(p);
      while (!tmp.equals(ONE)) {
        tmp = tmp.multiply(tmp).mod(p);
        i++;
      }  
      b = c;
      for (j = 0; j < m - i - 1; j++) {
        b = b.multiply(b).mod(p);
      }
      r = r.multiply(b).mod(p);
      t = t.multiply(b).multiply(b).mod(p);
      c = b.multiply(b).mod(p);
      m = i;
    }
    return r;
  }

  public static BigInteger modPow(BigInteger a, BigInteger n, BigInteger p)
  {
    BigInteger result = ONE;
    while (!n.equals(ZERO)) {
      if (n.testBit(0)) {
        result = result.multiply(a).mod(p);
        n = n.subtract(ONE);
      }
      a = a.multiply(a).mod(p);
      n = n.divide(TWO);
    }
    return result;
  }

  public static BigInteger selectZ(BigInteger p)
  {
    int l = p.bitLength();
    BigInteger a = new BigInteger(l, rand);
    while (isQR(a, p)) {
      a = new BigInteger(l, rand);
    }
    return a.mod(p);
  }

  public static boolean isQR(BigInteger a, BigInteger p) {
    if (a.mod(p).equals(ZERO)) return true;
    return modPow(a, (p.subtract(ONE)).divide(TWO), p).equals(ONE);
  }

  public static void main(String[] args)
  {
    System.out.println(modSqrt(new BigInteger("9028387512132124560753"), new BigInteger("13")));
  }
}
