import java.math.*;
import java.util.*;

public class QuadSieve
{
  public static BigInteger ONE = BigInteger.ONE;
  public static BigInteger ZERO = BigInteger.ZERO;

  public static ArrayList<BigInteger> initValues(BigInteger n, BigInteger start, int numValues)
  {
    BigInteger t = start;
    ArrayList<BigInteger> result = new ArrayList<BigInteger>();
    for (int i = 0; i < numValues; i++) {
      result.add(t.multiply(t).subtract(n));
      t = t.add(ONE);      
    }      
    return result;
  }

  public static ArrayList<BigInteger> sieve(BigInteger n, BigInteger start, int sieveSize,
                                            ArrayList<BigInteger> factorBase)
  {
    ArrayList<BigInteger> sieveValues = initValues(n, start, sieveSize);
    ArrayList<BigInteger> sieveCopy = new ArrayList<BigInteger>(sieveValues.size());
    for (int i = 0; i < sieveValues.size(); i++) sieveCopy.add(sieveValues.get(i));
    BigInteger end = start.add(new BigInteger((sieveSize - 1) + ""));
    for (BigInteger p : factorBase) {
      //System.out.println("Prime: " + p);
      BigInteger x = ModSqrt.modSqrt(n, p);
      if (x == null) {
        continue;
      }
      //System.out.println("Square Root: " + x);
      BigInteger otherX = p.subtract(x);

      BigInteger kstart = ((start.subtract(x)).divide(p)).add(ONE);
      BigInteger kend = ((end.subtract(x)).divide(p));
      for (BigInteger k = kstart; k.compareTo(kend) <= 0; k = k.add(ONE)) {
        int idx = (x.add(p.multiply(k)).subtract(start)).intValue();
        while (sieveCopy.get(idx).mod(p).equals(ZERO)) {
          sieveCopy.set(idx, sieveCopy.get(idx).divide(p));
        } 
      }

      if (x.equals(otherX)) continue;

      kstart = ((start.subtract(otherX)).divide(p)).add(ONE);
      kend = ((end.subtract(otherX)).divide(p));
      for (BigInteger k = kstart; k.compareTo(kend) <= 0; k = k.add(ONE)) {
        int idx = (otherX.add(p.multiply(k)).subtract(start)).intValue();
        while (sieveCopy.get(idx).mod(p).equals(ZERO)) {
          sieveCopy.set(idx, sieveCopy.get(idx).divide(p));
        } 
      }
    }
    ArrayList<BigInteger> result = new ArrayList<BigInteger>();
    for (int i = 0; i < sieveCopy.size(); i++) {
      if (sieveCopy.get(i).equals(ONE)) {
        result.add(sieveValues.get(i));
      }
    }
    return result;
  }

  public static BigInteger sqrt(BigInteger n) {
    BigInteger a = ONE;
    BigInteger b = new BigInteger(n.shiftRight(5).add(new BigInteger("8")).toString());
    while(b.compareTo(a) >= 0) {
      BigInteger mid = new BigInteger(a.add(b).shiftRight(1).toString());
      if(mid.multiply(mid).compareTo(n) > 0) b = mid.subtract(BigInteger.ONE);
      else a = mid.add(ONE);
    }
    return a.subtract(ONE);
  }

  public static void main(String[] args)
  {
    Random rnd = new Random();
    BigInteger n = new BigInteger("124335690703145647121956284919");
    BigInteger start = sqrt(n);
    int count = 1550000;
    ArrayList<BigInteger> factorBase = PrimeGen.generate(n, new BigInteger("50000"));
    System.out.println("Factor Base has size: " + factorBase.size());
    ArrayList<BigInteger> result = sieve(n, start, count, factorBase);
    int [][] matrix = LinearAlgebra.buildMatrix(result, factorBase);
    int w = 0;
    for (int i = 0; i < matrix.length; i++) {
      for (int j = 0; j < matrix[0].length; j++) {
        w += matrix[i][j];
      }
    }
    System.out.println("Result has size: " + result.size());
    System.out.println("Matrix has size: " + (matrix.length * matrix[0].length));
    System.out.println("Matrix has weight: " + w);
  }
}
